using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace LD10Tanks
{
    /// <summary>
    /// Mathematical 3D field that can be evaluated for inside/outside queries.
    /// </summary>
    public class EvalField
    {
        public struct SphereInfo
        {
            public Vector3 pos;
            public float radius;
            public float radius2;
        }

        public struct BoxInfo
        {
            public Vector3 center;
            public Vector3 axis_x;
            public Vector3 axis_y;
            public Vector3 axis_z;
            public Vector3 size;
        }

        public struct CylinderInfo
        {
            public Vector3 pt1;
            public Vector3 pt2;
            public Vector3 dir; // Not normalized; pt2-pt1
            public float len2; // The length of dir ^ 2
            public float radius1;
            public float radius2;
            public bool capsule;
        }

        private List<SphereInfo> m_spheres = new List<SphereInfo>();
        private List<BoxInfo> m_boxes = new List<BoxInfo>();
        private List<CylinderInfo> m_cylinders = new List<CylinderInfo>();

        private SphereGrid m_sphere_grid;

        public EvalField()
        {
        }

        public void Clear()
        {
            m_spheres.Clear();
            m_boxes.Clear();
            m_cylinders.Clear();

            if (m_sphere_grid != null)
                m_sphere_grid.Clear();
        }

        public void CopyFrom(EvalField field)
        {
            Clear();

            m_spheres.AddRange(field.m_spheres);
            m_boxes.AddRange(field.m_boxes);
            m_cylinders.AddRange(field.m_cylinders);

            if (m_sphere_grid != null)
            {
                for (int i = 0; i < m_spheres.Count; ++i)
                {
                    m_sphere_grid.AddSphere(m_spheres[i].pos, m_spheres[i].radius);
                }
            }
        }

        public void InitSphereGrid(Vector2 min_xz, Vector2 max_xz, float cell_size)
        {
            m_sphere_grid = new SphereGrid(min_xz, max_xz, cell_size);

            // Add existing spheres to the grid
            for (int i = 0; i < m_spheres.Count; ++i)
            {
                m_sphere_grid.AddSphere(m_spheres[i].pos, m_spheres[i].radius);
            }
        }

        public void AddSphere(Vector3 pos, float radius)
        {
            SphereInfo info;
            info.pos = pos;
            info.radius = radius;
            info.radius2 = radius * radius;

            m_spheres.Add(info);

            if (m_sphere_grid != null)
                m_sphere_grid.AddSphere(pos, radius);
        }

        public void AddBox(BoundingBox box)
        {
            BoxInfo info;
            info.center = (box.Min + box.Max) * 0.5f;
            info.axis_x = new Vector3(1, 0, 0);
            info.axis_y = new Vector3(0, 1, 0);
            info.axis_z = new Vector3(0, 0, 1);
            info.size = (box.Max - box.Min) * 0.5f;

            m_boxes.Add(info);
        }

        public void AddBox(Vector3 center, Vector3 dir, Vector3 up, float size_dir, float size_up, float size_side)
        {
            BoxInfo info;
            info.center = center;
            info.axis_x = dir;
            info.axis_x.Normalize();
            info.axis_y = Vector3.Cross(dir, up);
            info.axis_y.Normalize();
            info.axis_z = Vector3.Cross(info.axis_x, info.axis_y);
            info.size = new Vector3(size_dir * 0.5f, size_side * 0.5f, size_up * 0.5f);

            m_boxes.Add(info);
        }

        public void AddCylinder(Vector3 pt1, Vector3 pt2, float radius1, float radius2)
        {
            AddCylinder(pt1, pt2, radius1, radius2, false);
        }

        public void AddCylinder(Vector3 pt1, Vector3 pt2, float radius1, float radius2, bool capsule)
        {
            CylinderInfo info;
            info.pt1 = pt1;
            info.pt2 = pt2;
            info.dir = pt2 - pt1;
            info.len2 = info.dir.LengthSquared();
            info.radius1 = radius1;
            info.radius2 = radius2;
            info.capsule = capsule;

            m_cylinders.Add(info);
        }

        public bool IsInside(ref Vector3 pt)
        {
            // Test all spheres
            if (m_sphere_grid != null)
            {
                if (m_sphere_grid.IsPointInside(pt))
                    return true;
            }
            else
            { // Brute test otherwise
                for (int i = 0; i < m_spheres.Count; ++i)
                {
                    Vector3 tmp = pt - m_spheres[i].pos;
                    if (tmp.LengthSquared() <= m_spheres[i].radius2)
                        return true;
                }
            }

            // Test all boxes
            for (int i = 0; i < m_boxes.Count; ++i)
            {
                Vector3 tmp = pt - m_boxes[i].center;
                float x_dist = Math.Abs(Vector3.Dot(tmp, m_boxes[i].axis_x));
                float y_dist = Math.Abs(Vector3.Dot(tmp, m_boxes[i].axis_y));
                float z_dist = Math.Abs(Vector3.Dot(tmp, m_boxes[i].axis_z));

                if (x_dist <= m_boxes[i].size.X &&
                    y_dist <= m_boxes[i].size.Y &&
                    z_dist <= m_boxes[i].size.Z)
                {
                    return true;
                }
            }

            // Test all cylinders
            for (int i = 0; i < m_cylinders.Count; ++i)
            {
                float t = -Vector3.Dot(m_cylinders[i].pt1 - pt, m_cylinders[i].dir) / m_cylinders[i].len2;

                if (t < 0 || t > 1)
                {
                    if (!m_cylinders[i].capsule)
                        continue;

                    // Get the required distance
                    float req_dist;
                    Vector3 test_pt;

                    if (t < 0)
                    {
                        req_dist = m_cylinders[i].radius1;
                        test_pt = m_cylinders[i].pt1;
                    }
                    else
                    {
                        req_dist = m_cylinders[i].radius2;
                        test_pt = m_cylinders[i].pt2;
                    }

                    // Spherical test
                    if ((pt - test_pt).LengthSquared() <= (req_dist * req_dist))
                        return true;
                }
                else
                { // Inside the segment
                    // Get the required distance
                    float req_dist = (1.0f - t) * m_cylinders[i].radius1 + t * m_cylinders[i].radius2;

                    // Spherical test
                    Vector3 test_pt = m_cylinders[i].pt1 + t * m_cylinders[i].dir;
                    if ((pt - test_pt).LengthSquared() <= (req_dist * req_dist))
                        return true;
                }
            }

            return false;
        }
    }
}
