using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace LD10Tanks
{
    /// <summary>
    /// A simple 2D grid containing spheres. This is for point-inside-sphere queries.
    /// </summary>
    public class SphereGrid
    {
        private struct Item
        {
            public Vector3 center;
            public float radius2;
        }

        private struct Cell
        {
            public Item[] list;
            public int size;
        }

        private Cell[] m_cells;
        private float m_cell_size;
        private int m_cells_x;
        private int m_cells_y;
        private Vector2 m_grid_min;

        public SphereGrid(Vector2 min_xz, Vector2 max_xz, float cell_size)
        {
            m_cell_size = cell_size;
            m_cells_x = (int)((max_xz.X - min_xz.X) / cell_size) + 1;
            m_cells_y = (int)((max_xz.Y - min_xz.Y) / cell_size) + 1;

            m_cells = new Cell[m_cells_x * m_cells_y];
            for (int i = 0; i < m_cells.Length; ++i)
            {
                m_cells[i].list = new Item[8];
            }

            m_grid_min = min_xz;
        }

        public void Clear()
        {
            for (int i = 0; i < m_cells.Length; ++i)
            {
                m_cells[i].size = 0;
            }
        }

        public bool IsPointInside(Vector3 pt)
        {
            Point cell = GetPointCell(ref pt);
            if (cell.X < 0 || cell.Y < 0 || cell.X >= m_cells_x || cell.Y >= m_cells_y)
                return false;

            int idx = cell.X + cell.Y * m_cells_x;
            Item[] lst = m_cells[idx].list;
            int amt = m_cells[idx].size;

            for (int i = 0; i < amt; ++i)
            {
                Vector3 dir = pt - lst[i].center;
                if (dir.LengthSquared() <= lst[i].radius2)
                    return true;
            }
            return false;
        }

        public void AddSphere(Vector3 center, float radius)
        {
            Vector2 min = new Vector2(center.X - radius, center.Z - radius);
            Vector2 max = new Vector2(center.X + radius, center.Z + radius);
            Point min_pt = GetPointCell(min);
            Point max_pt = GetPointCell(max);

            for (int y = min_pt.Y; y <= max_pt.Y; ++y)
            {
                if (y < 0 || y >= m_cells_y)
                    continue;

                for (int x = min_pt.X; x <= max_pt.X; ++x)
                {
                    if (x < 0 || x >= m_cells_x)
                        continue;

                    int idx = x + y * m_cells_x;
                    Item[] lst = m_cells[idx].list;
                    int pos = m_cells[idx].size;

                    if (pos >= lst.Length)
                    { // Need to make a larger list
                        Item[] new_lst = new Item[lst.Length * 2];
                        Array.Copy(lst, new_lst, lst.Length);

                        lst = new_lst;
                        m_cells[idx].list = lst;
                    }

                    lst[pos].center = center;
                    lst[pos].radius2 = radius * radius;
                    m_cells[idx].size = pos + 1;
                }
            }
        }

        private Point GetPointCell(Vector2 pt)
        {
            return new Point(
                (int)((pt.X - m_grid_min.X) / m_cell_size),
                (int)((pt.Y - m_grid_min.Y) / m_cell_size));
        }

        private Point GetPointCell(ref Vector3 pt)
        {
            return new Point(
                (int)((pt.X - m_grid_min.X) / m_cell_size),
                (int)((pt.Z - m_grid_min.Y) / m_cell_size));
        }
    }
}
